/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * hiTs program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* 
 * Note to regenerate the Makefile and configure script:
 * aclocal
 * autoconf
 * touch NEWS README AUTHORS ChangeLog
 * automake -a
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_FEATURES_H
/* this is defined to shut up gcc's warning about posix_memalign,
   but the code usually compiles even when this is not defined */
#define _GNU_SOURCE
#include <features.h>
#endif

#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <locale.h>

#if defined HAVE_LANGINFO_H
#include <langinfo.h>
#if !defined CODESET
/* on OpenBSD, CODESET doesn't seem to be defined - 
   we use 3, which should be US-ASCII, but it's not ideal... */
#define CODESET 3
#endif
#endif

#include "dbacl.h" /* make sure this is last */

/* global variables */

hash_bit_count_t default_max_hash_bits = 15;
hash_count_t default_max_tokens = (1<<15);

hash_bit_count_t default_max_grow_hash_bits = 16;
hash_count_t default_max_grow_tokens = (1<<16);

hash_bit_count_t decimation;

Dirichlet dirichlet;

Learner learner;
Category cat[MAX_CAT];
category_count_t cat_count = 0;
int filter[MAX_CAT];
category_count_t filter_count = 0;
/* the Regex array contains both regexes (first half) and antiregexes
   (second half) */
Regex re[MAX_RE];
regex_count_t regex_count = 0;
Empirical empirical;

extern token_order_t ngram_order; /* defaults to 1 */

/* for counting emails */
bool_t not_header; 
extern MBOX_State mbox;

/* for option processing */
extern char *optarg;
extern int optind, opterr, optopt;

options_t u_options = 0;
options_t m_options = 0;
char *title = "";
char *online = "";

int cmd = 0;
int exit_code = 0; /* default */

int overflow_warning = 0;
int digramic_overflow_warning = 0;
int skewed_constraints_warning = 0;

/* default value in case we don't have getpagesize() */
long system_pagesize = BUFSIZ;
void *in_iobuf = NULL; /* only used for input stream */
void *out_iobuf = NULL; /* used for all category/learner file operations */

/* tolerance for the error in divergence - this can be changed with -q.
   More accuracy means slower learning. This should be the same order
   of magnitude as the error due to digitization, but we tweak them
   to obtain timely output */
int quality = 0;
double qtol_alpha = 0.0001;
double qtol_div = 0.01;
double qtol_lam = 0.05;
double qtol_logz = 0.05;
bool_t qtol_multipass = 0;

/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

void usage(char **argv) {
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl [-vniNR] [-T type] -c CATEGORY [-c CATEGORY]...\n");
  fprintf(stderr, 
	  "      [-f KEEP]... [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      classifies FILE or STDIN using CATEGORY, optionally\n");
  fprintf(stderr, 
	  "      removing all lines which don't fit the category KEEP.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl [-vnirNDL] [-h size] [-T type] -l CATEGORY \n");
  fprintf(stderr, 
	  "      [-g regex]... [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      builds a maximum entropy model from the words in FILE or STDIN\n");
  fprintf(stderr, 
	  "      or concatenated regex submatches if using the -g option.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl -V\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      prints program version.\n");
}


/***********************************************************
 * CATEGORY FUNCTIONS                                      *
 ***********************************************************/

/* note: don't forget to flush after each line */
void line_score_single_category(char *textbuf) {
    score_t score = 0.0;

    if( *textbuf ) {
      if( u_options & (1<<U_OPTION_POSTERIOR) ) {
	/* score = empirical_entropy(); */
	fprintf(stdout, "%s", textbuf);
      }
      if( u_options & (1<<U_OPTION_SCORES) ) {
	if( u_options & (1<<U_OPTION_VERBOSE) ) {
	  score = -(cat[0].score/cat[0].complexity)/log(2.0);
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " * %-4d %s", 
		  cat[0].filename, score, 
		  cat[0].complexity, textbuf);
	} else {
	  fprintf(stdout, "%s %7.2" FMT_printf_score_t " %s", 
		  cat[0].filename, -cat[0].score, textbuf); 
	}
      }
      fflush(stdout);
    }
    cat[0].score = 0.0;
    cat[0].complexity = 0;

    if( m_options & (1<<M_OPTION_CALCENTROPY) ) {
      clear_empirical(&empirical);
    }
}

/* note: don't forget to flush after each line */
void line_score_multiple_categories(char *textbuf) {
  category_count_t i;
  int map;
  score_t c, cmax;

  /* find MAP */
  cmax = cat[0].score; 
  map = 0;
  for(i = 1; i < cat_count; i++) {
    if(cmax < cat[i].score) {
      cmax = cat[i].score;
      map = i;
    }
  }

  if(u_options & (1<<U_OPTION_POSTERIOR) ) {
    /* compute probabilities given exclusive choices */
    c = 0.0;
    for(i = 0; i < cat_count; i++) {
      c += exp((cat[i].score - cmax));
    }
      
    if( *textbuf ) {
      for(i = 0; i < cat_count; i++) {
	fprintf(stdout, "%s %6.2" FMT_printf_score_t "%% ", 
		cat[i].filename, 100 * exp((cat[i].score - cmax))/c);
      }
      fprintf(stdout, "%s", textbuf);
      fflush(stdout);
    }
  } else if( u_options & (1<<U_OPTION_SCORES) ) {
    /* display normalized divergence scores */
    if( *textbuf ) {
      for(i = 0; i < cat_count; i++) {
	if( u_options & (1<<U_OPTION_VERBOSE) ) {
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " * %-4d ", 
		  cat[i].filename, 
		  -cat[i].score/cat[i].complexity, 
		  cat[i].complexity);
	} else {
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " ", 
		  cat[i].filename, -cat[i].score);
	}
      }
      fprintf(stdout, "%s", textbuf);
      fflush(stdout);
    }
  } else {
    /* prune the text which doesn't fit */
    for(i = 0; i < filter_count; i++) {
      if( cat[map].score <= cat[filter[i]].score ) {
	fprintf(stdout, "%s", textbuf);
	fflush(stdout);
	break;
      }
    }
  }
    
  /* clean up for next line */
  for(i = 0; i < cat_count; i++) {
    cat[i].score = 0.0;
    cat[i].complexity = 0;
  }
  if( m_options & (1<<M_OPTION_CALCENTROPY) ) {
    clear_empirical(&empirical);
  }
}

void score_single_category() {

  if( u_options & (1<<U_OPTION_POSTERIOR) ) {
    /* this is blindingly obvious, but needs to be said */
    fprintf(stdout, "%s 100%%\n", cat[0].filename);
  }

  if( u_options & (1<<U_OPTION_APPEND) ) {
    fprintf(stdout, "\nscores ");
  }

  if( u_options & (1<<U_OPTION_SCORES) ) {
    if( u_options & (1<<U_OPTION_VERBOSE) ) {
      fprintf(stdout, "cross_entropy %.2" FMT_printf_score_t \
                      " bits complexity %d\n", 
	      -(cat[0].score/cat[0].complexity) / log(2.0), cat[0].complexity);
    } else {
      fprintf(stdout, "score %.2" FMT_printf_score_t "\n", -cat[0].score);
    }

    if( u_options & (1<<U_OPTION_APPEND) ) {
      if( cat[0].model_num_docs > 0 ) {
	fprintf(stdout, "mean_complexity %s %5.2" FMT_printf_score_t "\n", 
		cat[0].filename, 
		(double)cat[0].model_full_token_count/cat[0].model_num_docs);
      }
    }

  }

  exit_code = 1;
}

void score_multiple_categories() {
  bool_t no_title;
  category_count_t i;
  score_t c, cmax;

  /* find MAP */
  cmax = cat[0].score; 
  exit_code = 0;
  for(i = 1; i < cat_count; i++) {
    if(cmax < cat[i].score) {
      cmax = cat[i].score;
      exit_code = i;
    }
  }

  if( u_options & (1<<U_OPTION_POSTERIOR) ) {
    if( u_options & (1<<U_OPTION_APPEND) ) {
      fprintf(stdout, "\nprobabilities ");
    }
    /* here we compute probabilities given exclusive choices */
    c = 0.0;
    for(i = 0; i < cat_count; i++) {
      c += exp((cat[i].score - cmax));
    }

    for(i = 0; i < cat_count; i++) {
      fprintf(stdout, "%s %5.2" FMT_printf_score_t "%% ", 
	      cat[i].filename, 100 * exp((cat[i].score - cmax))/c);
    }
    fprintf(stdout, "\n");
  } else if( u_options & (1<<U_OPTION_SCORES) ) {
    if( u_options & (1<<U_OPTION_APPEND) ) {
      fprintf(stdout, "\nscores ");
    }
    /* display logarithmic score */
    for(i = 0; i < cat_count; i++) {
      if( u_options & (1<<U_OPTION_VERBOSE) ) {
	fprintf(stdout, "%s %5.2" FMT_printf_score_t " * %-d ", 
		cat[i].filename, (-cat[i].score/cat[i].complexity)/log(2.0), 
		cat[i].complexity);
      } else {
	fprintf(stdout, "%s %5.2" FMT_printf_score_t " ", 
		cat[i].filename, -cat[i].score/log(2.0));
      }
    }
    fprintf(stdout, "\n");

    if( u_options & (1<<U_OPTION_APPEND) ) {
      no_title = 1;
      for(i = 0; i < cat_count; i++) {
	if( cat[i].model_num_docs > 0 ) {
	  if( no_title ) {
	    fprintf(stdout, "mean_complexity ");
	    no_title = 0;
	  }
	  fprintf(stdout, "%s %5.2" FMT_printf_score_t " ", 
		  cat[i].filename, 
		  (double)cat[i].model_full_token_count/cat[i].model_num_docs);
	}
      }
      if( !no_title ) { fprintf(stdout, "\n"); }
    }


  } else if( u_options & (1<<U_OPTION_VERBOSE) ) {
    if( u_options & (1<<U_OPTION_APPEND) ) {
      fprintf(stdout, "\ncategory ");
    }
    fprintf(stdout, "%s\n", cat[exit_code].filename); 
  }

  exit_code++; /* make number between 1 and cat_count+1 */
}

/***********************************************************
 * FILE MANAGEMENT FUNCTIONS                               *
 ***********************************************************/

bool_t check_magic_write(char *path, char *magic, int len) {
  FILE *output;
  char buf[MAGIC_BUFSIZE];

  if( (output = fopen(path, "r")) ) {
    /* output file exists already */
    if( (!feof(output) && !fgets(buf, MAGIC_BUFSIZE, output)) ||
	(strncmp(buf, magic, len) != 0) ) {
      fprintf(stderr,
	      "error: the file %s is already used for something, "
	      "use another filename. Nothing written.\n", path);
      fclose(output);
      return 0;
    } else {
      /* it's an existing category file */
      fclose(output);
    }
  }
  return 1;
}

/* remember to free the tmpname when done */

/* the standard tmpfile() call doesn't tell the filename,
   the standard mkstemp() is unreliable,
   and the moronic unix system doesn't remember filenames on open() */
FILE *mytmpfile(const char *template, char **tmpname) {
  int i, fd, l;
  FILE *result = NULL;
  l = strlen(template);
  *tmpname = (char *)malloc(sizeof(char)*(l + 3));
  if( *tmpname ) {
    strcpy(*tmpname, template);
    for(i = 0; i < 100; i++) {
      snprintf(*tmpname + l, 3, ".%d", i);
      /* this may have problems on NFS systems? */
      fd = open(*tmpname, O_CREAT|O_EXCL|O_WRONLY, 0440);
      if( fd != -1 ) {
	result = fdopen(fd, "wb");
	/* no need to close fd */
	break;
      }
    }
  }
  return result;
}

/* writes the learner to a file for easily readable category */
/* the category file is first constructed as a temporary file,
   then renamed if no problems occured. Because renames are 
   atomic on POSIX, this guarantees that a loaded category file
   used for classifications is never internally corrupt */
error_code_t save_learner() {
  alphabet_size_t i, j;
  hash_count_t t;
  regex_count_t c;
  FILE *output;
  char *tempname;

  bool_t ok;
  int n;
  c_item ci;
  
  char scratchbuf[MAGIC_BUFSIZE];
  char smb[MAX_SUBMATCH+1];
  token_order_t s;
  char *p;

#if defined DIGITIZE_DIGRAMS
  digitized_weight_t shval;
#else
  weight_t shval;
#endif

  if( u_options & (1<<U_OPTION_VERBOSE) ) {
    fprintf(stdout, "saving category to file %s\n", learner.filename);
  }
  
  /* don't overwrite data files */
  if( !check_magic_write(learner.filename, MAGIC1, 10) ) {
    exit(0);
  }
  
  /* this keeps track to see if writing is successful, 
     it's not foolproof, but probably good enough */
  ok = 1; 

  if( (output = mytmpfile(learner.filename, &tempname)) ) {
    
    if( out_iobuf ) {
      setvbuf(output, out_iobuf, _IOFBF, BUFFER_MAG * system_pagesize);
    }

    /* print out standard headers */
    ok = ok && 
      (0 < fprintf(output, MAGIC1, learner.filename, 
	       (m_options & (1<<M_OPTION_REFMODEL)) ? "(ref)" : ""));
    ok = ok &&
      (0 < fprintf(output, 
		   MAGIC2_o, learner.divergence, learner.logZ, learner.max_order,
		   (m_options & (1<<M_OPTION_MULTINOMIAL)) ? "multinomial" : "hierarchical" ));
    ok = ok &&
      (0 < fprintf(output, MAGIC3, 
		   (short int)learner.max_hash_bits, 
		   (long int)learner.full_token_count, 
		   (long int)learner.unique_token_count,
		   (long int)learner.num_docs));

    /* print out any regexes we might need */
    for(c = 0; c < regex_count; c++) {
      /* write the bitmap */
      for(p = smb, s = 1; s <= MAX_SUBMATCH; s++) {
	if( re[c].submatches & (1<<s) ) {
	  *p++ = s + '0';
	}
      }
      *p = '\0';

#if defined HAVE_LIBBOOST_REGEX
      /* does this work? */
      ok = ok && 
	(0 < fprintf(output, MAGIC5_wo, re[c].string, smb));
#else
      ok = ok && 
	(0 < fprintf(output, MAGIC5_o, re[c].string, smb));
#endif
    }

    /* print options */
    ok = ok &&
      (0 < fprintf(output, MAGIC4_o, m_options, 
		   print_model_options(m_options, scratchbuf)));

    ok = ok &&
      (0 < fprintf(output, MAGIC6)); 

    /* end of readable stuff */
    if( ok ) {
      /* now write arrays: this is far(!) from efficient, but the bits
	 we need are not nicely tucked into a block of memory, and due
	 to the sizes involved when dealing with possibly millions of tokens, we
	 can't afford to malloc a brand new block just to pack everything 
	 nicely. At least, the file buffering should mitigate the performance
	 hit, and we prefer to learn slowly and classify fast. */

      /* character frequencies */
      for(i = 0; i < ASIZE; i++) {
	for(j = 0; j < ASIZE; j++) {
	  shval = PACK_DIGRAMS(learner.dig[i][j]);
	  for(n = 0; n < 1; ) {
	    if( 0 > (n = fwrite(&shval, SIZEOF_DIGRAMS, 1, output)) ) {
	      ok = 0;
	      goto skip_remaining;
	    } 
	  }	    
	}
      }

#if defined HAVE_MADVISE
      madvise(learner.hash, sizeof(l_item) * learner.max_tokens, 
	      MADV_SEQUENTIAL);
#endif

      /* token/feature weights */
      for(t = 0; t < learner.max_tokens; t++) {
	/* write each element so that it's easy to read back in a c_item array */
	SET(ci.id,learner.hash[t].id);
#if defined DIGITIZE_LAMBDA
	if( fabs((learner.hash[t].lam)) >= (1<<(16 - DIG_FACTOR)) ) {
	  fprintf(stderr,
		  "warning: digitized an extreme lambda value (%f) - results may be unusable\n", learner.hash[t].lam);
	  m_options |= (1<<M_OPTION_WARNING_BAD);
	  process_pending_signal(NULL);
	} 
	ci.lam = PACK_LWEIGHTS(learner.hash[t].lam);
#else
	ci.lam = (learner.hash[t].lam);
#endif
	for(n = 0; n < 1; ) {
	  if( 0 > (n = fwrite(&ci, sizeof(ci), 1, output)) ) {
	    ok = 0;
	    goto skip_remaining;
	  } 
	}
      }
    }

  skip_remaining:

    fclose(output);

    /* the rename is atomic on posix */
    if( !ok || (rename(tempname, learner.filename) < 0) ) { 
      fprintf(stderr, "error: due to a potential file corruption, category %s was not updated\n", 
	      learner.filename);
      unlink(tempname);
    }
    free(tempname);
  } else {
    fprintf(stderr, "error: cannot open tempfile for writing %s\n", 
	    learner.filename);
    return 0;
  }


  return 1;
}

/***********************************************************
 * LEARNER FUNCTIONS                                       *
 ***********************************************************/

void read_online_learner_struct(char *path) {
  FILE *input;
  char buf[BUFSIZ+1]; /* must be greater than MAGIC_BUFSIZE */
  bool_t ok = 1;
  char *new_filename;
  FILE *new_tmp;
  size_t t, j, n;

  if( (input = fopen(path, "r")) ) {

    if( out_iobuf ) {
      setvbuf(input, out_iobuf, _IOFBF, BUFFER_MAG * system_pagesize);
    }

    if( !fgets(buf, MAGIC_BUFSIZE, input) ||
	(strncmp(buf, MAGIC_ONLINE, strlen(MAGIC_ONLINE)) != 0) ) {
	if( strncmp(buf, MAGIC_ONLINE, 35) == 0 ) {
	  fprintf(stderr,
		  "warning: the file %s has the wrong version number, it will be ignored\n", 
		  path);
	} else {
	  fprintf(stderr,
		  "warning: the file %s is not a dbacl online memory dump, it will be ignored\n", 
		  path);
	}
    } else {
      /* from here on, if a problem arises then learner is hosed, so we exit */
      ok = 1;
      /* first read learner */
      new_filename = learner.filename;
      new_tmp = learner.tmp;
      if( learner.hash ) { free(learner.hash); }
      for(n = 0; n < 1; ) {
	if( 0 > (n = fread(&learner, sizeof(learner), 1, input)) ) {
	  ok = 0;
	  goto skip_read_online;
	}
      }
      learner.filename = new_filename;
      learner.tmp = new_tmp;

      learner.hash = (l_item *)malloc(learner.max_tokens * sizeof(l_item));
      if( !learner.hash ) {
	  ok = 0;
	goto skip_read_online;
      }

#if defined HAVE_MADVISE
      madvise(learner.hash, sizeof(l_item) * learner.max_tokens, 
	      MADV_SEQUENTIAL);
#endif

      for(n = j = 0; n < learner.max_tokens; n += j) {
	if( 0 > (j = fread(&learner.hash[n], sizeof(l_item), 
			   learner.max_tokens - n, input)) ) {
	  ok = 0;
	  goto skip_read_online;
	}
      }
      /* next read empirical */
      if( empirical.hash ) { free(empirical.hash); }
      for(n = 0; n < 1; ) {
	if( 0 > (n = fread(&empirical, sizeof(empirical), 1, input)) ) {
	  ok = 0;
	  goto skip_read_online;
	}
      }
      empirical.hash = (h_item *)malloc(empirical.max_tokens * sizeof(h_item));
      if( !empirical.hash ) {
	  ok = 0;
	goto skip_read_online;
      }

#if defined HAVE_MADVISE
      madvise(empirical.hash, sizeof(h_item) * empirical.max_tokens, 
	      MADV_SEQUENTIAL);
#endif

      for(n = j = 0; n < empirical.max_tokens; n += j) {
	if( 0 > (j = fread(&empirical.hash[n], sizeof(h_item), 
			   empirical.max_tokens - n, input)) ) {
	  ok = 0;
	  goto skip_read_online;
	}
      }
      /* last, copy the token strings into the temporary file */
      n = 0;
      while( (n >= 0) && !feof(input) ) {
	n = fread(buf, 1, BUFSIZ, input);
	if( n > 0 ) {
	  for(t = j = 0; t < n; t += j) {
	    if( 0 > (j = fwrite(buf + t, 1, n - t, learner.tmp)) ) {
	      ok = 0;
	      goto skip_read_online;
	    }
	  }
	}
      }
      
    }
  skip_read_online:
    fclose(input);
    if( !ok ) {
      fprintf(stderr, "error: there was an unrecoverable problem reading %s\n",
	      path);
      exit(0);
    }
  }
}

/* this is a straight memory dump  */
void write_online_learner_struct(char *path) {
  FILE *output;
  char *tempname;
  bool_t ok;
  char buf[BUFSIZ+1];
  size_t t, j, n;

  if( !check_magic_write(path, MAGIC_ONLINE, strlen(MAGIC_ONLINE)) ) {
    /* we simply ignore this. Note that check_magic_write() already
       notifies the user */
    return;
  }

  if( (output = mytmpfile(path, &tempname)) ) {

    if( out_iobuf ) {
      setvbuf(output, out_iobuf, _IOFBF, BUFFER_MAG * system_pagesize);
    }

    ok = 1;
    ok = ok &&
      (0 < fprintf(output, MAGIC_ONLINE));
    /* write learner */
    ok = ok &&
      (0 < fwrite(&learner, sizeof(learner), 1, output));
    for(t = j = 0; t < learner.max_tokens; t += j) {
      if( 0 > (j = fwrite(&learner.hash[t], sizeof(l_item), 
			  learner.max_tokens - t, output)) ) {
	ok = 0;
	goto skip_write_online;
      }
    }
    /* write empirical */
    ok = ok &&
      (0 < fwrite(&empirical, sizeof(empirical), 1, output));
    for(t = j = 0; t < empirical.max_tokens; t += j) {
      if( 0 > (j = fwrite(&empirical.hash[t], sizeof(l_item), 
			  empirical.max_tokens - t, output)) ) {
	ok = 0;
	goto skip_write_online;
      }
    }
    /* write temporary tokens */
    rewind(learner.tmp);
    n = 0;
    while( (n >= 0) && !feof(learner.tmp) ) {
      n = fread(buf, 1, BUFSIZ, learner.tmp);
      if( n > 0 ) {
	for(t = j = 0; t < n; t += j) {
	  if( 0 > (j = fwrite(buf + t, 1, n - t, output)) ) {
	    ok = 0;
	    goto skip_write_online;
	  }
	}
      }
    }

  skip_write_online:
    fclose(output);

    /* the rename is atomic on posix */
    if( !ok || (rename(tempname, path) < 0) ) {
      fprintf(stderr, "error: due to a potential file corruption, %s was not updated\n", 
	      path);
      unlink(tempname);
    }
    free(tempname);
  }
}

void reset_mbox_messages() {
  not_header = 1;
  learner.num_docs = 0;
  learner.skip_doc = 0;
}

void count_mbox_messages(char *textbuf) {

  switch(mbox.state) {
  case msHEADER:
    if(not_header) {
      if( u_options & (1<<U_OPTION_DECIMATE) ) {
	learner.skip_doc = ( rand() > (RAND_MAX>>decimation) );
      } 
      if( !learner.skip_doc ) {
	learner.num_docs++;
      }
    }
    not_header = 0;
    break;
  default:
    not_header = 1;
    break;
  }

}


l_item *find_in_learner(hash_value_t id) {
    register l_item *i, *loop;
    /* start at id */
    i = loop = &learner.hash[id & (learner.max_tokens - 1)];

    while( FILLEDP(i) ) {
	if( EQUALP(i->id,id) ) {
	    return i; /* found id */
	} else {
	    i++; /* not found */
	    /* wrap around */
	    i = (i >= &learner.hash[learner.max_tokens]) ? learner.hash : i; 
	    if( i == loop ) {
		return NULL; /* when hash table is full */
	    }
	}
    }

    /* empty slot, so not found */

    return i; 
}

/* returns true if the hash could be grown, false otherwise.
   When the hash is grown, the old values must be redistributed.
   To do this, we reuse the l_item.lam member as a flag, since this
   is only needed later during minimization */
bool_t grow_learner_hash() {
  hash_count_t c, new_size;
  l_item *i, temp_item;

  if( !(u_options & (1<<U_OPTION_GROWHASH)) ) {
    return 0;
  } else {
    if( learner.max_hash_bits < default_max_grow_hash_bits ) {

      /* 
	 this warning is mostly useless and distracting

	 fprintf(stderr,
	 "warning: growing hash table, next time perhaps consider using -h %d\n",
	 learner.max_hash_bits + 1);

      */

      /* grow the memory around the hash */
      if( (i = realloc(learner.hash, 
		       sizeof(l_item) * 
		       (1<<(learner.max_hash_bits+1)))) == NULL ) {
	fprintf(stderr,
		"warning: sorry, failed to grow hash table\n");
	return 0;
      }

#if defined HAVE_MADVISE
      madvise(i, sizeof(l_item) * (1<<learner.max_hash_bits), MADV_SEQUENTIAL);
#endif

      learner.max_hash_bits++;
      /* realloc doesn't initialize the memory */
      memset(&((l_item *)i)[learner.max_tokens], 0, 
	     ((1<<learner.max_hash_bits) - learner.max_tokens) * 
	     sizeof(l_item));
      learner.hash = i; 
      
      /* now mark every used slot */
      for(c = 0; c < learner.max_tokens; c++) {
	if( FILLEDP(&learner.hash[c]) ) {
	  SETMARK(&learner.hash[c]);
	}
      }

#if defined HAVE_MADVISE
      madvise(i, sizeof(l_item) * (1<<learner.max_hash_bits), MADV_RANDOM);
#endif

      /* now relocate each marked slot and clear it */
      new_size = (1<<learner.max_hash_bits) - 1;
      for(c = 0; c < learner.max_tokens; c++) {
	while( MARKEDP(&learner.hash[c]) ) {
	  /* find where it belongs */
	  i = &learner.hash[learner.hash[c].id & new_size];
	  while( FILLEDP(i) && !MARKEDP(i) ) {
	    i++;
	    i = (i > &learner.hash[new_size]) ? learner.hash : i;
	  } /* guaranteed to exit since hash is larger than original */

	  /* clear the mark - this must happen after we look for i,
	   since it should be possible to find i == learner.hash[c] */
	  UNSETMARK(&learner.hash[c]); 

	  /* now refile */
	  if( i != &learner.hash[c] ) {
	    if( MARKEDP(i) ) {
	      /* swap */
	      memcpy(&temp_item, i, sizeof(l_item));
	      memcpy(i, &learner.hash[c], sizeof(l_item));
	      memcpy(&learner.hash[c], &temp_item, sizeof(l_item));
	    } else {
	      /* copy and clear */
	      memcpy(i, &learner.hash[c], sizeof(l_item));
	      memset(&learner.hash[c], 0, sizeof(l_item));
	    }
	  } 
	  /* now &learner.hash[c] is marked iff there was a swap */
	}
      }
      learner.max_tokens = (1<<learner.max_hash_bits);
    } else {
      u_options &= ~(1<<U_OPTION_GROWHASH); /* it's the law */
    }
  }
  return 1;
}


/* places the token in the global hash and writes the
   token to a temporary file for later, then updates
   the digram frequencies. Tokens have the format 
   DIAMOND t1 DIAMOND t2 ... tn DIAMOND CLASSEP class NUL */
void hash_word_and_learn(char *tok, token_type_t tt, regex_count_t re) {
  hash_value_t id;
  l_item *i;
  alphabet_size_t p,q;

  if( *(tok + EXTRA_TOKEN_LEN) ) { /* DIAMOND-DIAMOND-CLASSEP-class is empty */

    if( overflow_warning ) {
      return; /* for there be troubles ahead */
    }

    if( m_options & (1<<M_OPTION_MULTINOMIAL) ) { tt.order = 1; }

    if( u_options & (1<<U_OPTION_DECIMATE) ) {
      if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
	if( learner.skip_doc ) {
	  return;
	}
      } else {
	if( rand() > (RAND_MAX>>decimation) ) {
	  return;
	}
      }
    }

    id = (hash_value_t)hash((unsigned char *)tok, strlen(tok), 0);
    i = find_in_learner(id);

    if( i && !FILLEDP(i) &&
	((100 * learner.unique_token_count) >= 
	 (HASH_FULL * learner.max_tokens)) && grow_learner_hash() ) {
      i = find_in_learner(id);
      /* new i, go through all tests again */
    }

    if( i ) {

      if( FILLEDP(i) ) {

	/* redundant :-) */

      } else if( /* !FILLEDP(i) && */
		((100 * learner.unique_token_count) < 
		 (HASH_FULL * learner.max_tokens) ) ) {

	/* fill the hash and write to file */

	SET(i->id,id);

	if( learner.unique_token_count < K_TOKEN_COUNT_MAX )
	  { learner.unique_token_count++; } else { overflow_warning = 1; }

	i->typ = tt;

	/* order accounting */
	learner.max_order = (learner.max_order < i->typ.order) ? 
	  i->typ.order : learner.max_order;

	if( learner.fixed_order_unique_token_count[i->typ.order] < K_TOKEN_COUNT_MAX ) 
	  { learner.fixed_order_unique_token_count[i->typ.order]++; } else 
	    { overflow_warning = 1; }

     	if( (u_options & (1<<U_OPTION_DEBUG)) ) {
     	  fprintf(stdout, "match %s %d\n", tok, i->typ.order);
     	}

	/* now save token to temporary file */
	fputs(tok, learner.tmp);
	fputc(TOKENSEP, learner.tmp);
      }

      if( i->count < K_TOKEN_COUNT_MAX ) { 
	i->count++; 
      } else { 
	overflow_warning = 1; 
      }

      if( learner.full_token_count < K_TOKEN_COUNT_MAX )
	{ learner.full_token_count++; } else 
	  { /* this number is just cosmetic */ }

      if( learner.fixed_order_token_count[i->typ.order] < K_TOKEN_COUNT_MAX ) 
	{ learner.fixed_order_token_count[i->typ.order]++; } else 
	  { skewed_constraints_warning = 1; }

    }

    if( digramic_overflow_warning ) {
      return;
    }


    /* now update digram frequency counts */
    if( tt.order == 1 ) { /* count each token only once */
      p = *tok++;
      /* finally update character frequencies */
      while( *tok != EOTOKEN ) {
	q = (unsigned char)*tok;
	if( learner.dig[p][q] < K_TOKEN_COUNT_MAX ) 
	  { learner.dig[p][q]++; } else { digramic_overflow_warning = 1; }
	p = q;
	tok++;
      }

      if( digramic_overflow_warning ) {
	fprintf(stderr, 
		"warning: ran out of integers (too much data), "
		"reference measure may be skewed.\n");
	m_options |= (1<<M_OPTION_WARNING_BAD);
      }
    }

    if( overflow_warning ) {
      fprintf(stderr, 
	      "warning: ran out of integers (too much data), "
	      "results may be skewed.\n");
      m_options |= (1<<M_OPTION_WARNING_BAD);
    }
  }
}

/* initialize global learner object */
void init_learner() {
  alphabet_size_t i, j;
  regex_count_t c;
  token_order_t z;

  /* some constants */
  learner.max_tokens = default_max_tokens;
  learner.max_hash_bits = default_max_hash_bits;
  learner.full_token_count = 0;
  learner.unique_token_count = 0;
  learner.logZ = 0.0;
  learner.max_order = 0;

  /* init character frequencies */
  for(i = 0; i < ASIZE; i++) { 
    for(j = 0; j < ASIZE; j++) { 
      learner.dig[i][j] = 0.0;
    }
  }

  /* open temporary file for writing tokens */
  learner.tmp = tmpfile();
  learner.tmpiobuf = NULL;   /* can't reuse in_iobuf or out_iobuf */

  if( learner.tmp ) {
#if defined HAVE_POSIX_MEMALIGN
    /* buffer must exist until after fclose() */
    if( 0 != posix_memalign(&learner.tmpiobuf, system_pagesize, 
			    BUFFER_MAG * system_pagesize) ) {
      learner.tmpiobuf = NULL;
    }
#elif defined HAVE_MEMALIGN
    /* buffer can't be free()'d */
    learner.tmpiobuf = (void *)memalign(system_pagesize, 
					BUFFER_MAG * system_pagesize);
#elif defined HAVE_VALLOC
    /* buffer can't be free()'d */
    learner.tmpiobuf = (void *)valloc(BUFFER_MAG * system_pagesize);
#endif
    if( learner.tmpiobuf ) {
      setvbuf(learner.tmp, learner.tmpiobuf, _IOFBF, 
	      BUFFER_MAG * system_pagesize);
    }
  }

  /* allocate room for hash */
  learner.hash = calloc(learner.max_tokens, sizeof(l_item));
  if( !learner.hash ) {
    fprintf(stderr, 
	    "error: not enough memory? I couldn't allocate %li bytes\n", 
	    (sizeof(l_item) * ((long int)learner.max_tokens)));
    exit(0);
  }

#if defined HAVE_MADVISE
  madvise(learner.hash, sizeof(l_item) * learner.max_tokens, 
	  MADV_SEQUENTIAL);
#endif

  for(c = 0; c < MAX_RE + 1; c++) {
    learner.regex_token_count[c] = 0;
  }

  for(z = 0; z < MAX_SUBMATCH; z++) {
    learner.fixed_order_token_count[z] = 0;
    learner.fixed_order_unique_token_count[z] = 0;
  }

  if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
    reset_mbox_messages();
  }

  if( *online ) {
    read_online_learner_struct(online);
  }

#if defined HAVE_MADVISE
  madvise(learner.hash, sizeof(l_item) * learner.max_tokens, 
	  MADV_RANDOM);
#endif
}

/* calculates the most probable Dirichlet parameters 
   given digram counts. Method described in MacKay and Peto (1995)
   Since we don't count the DIAMOND-DIAMOND digram, including the prior will 
   overestimate the DIAMOND-DIAMOND transition probability. However, for this 
   transition as well as the others, the Dirichlet will have practically
   no effect when there's enough data in the corpus.

   HOWEVER: 
   * A property of the MP posterior estimates is that if character j
   * never appears as a transition *target*, then the corresponding
   * u[j] entry is null. Sometimes, this makes sense, as j isn't part
   * of the "vocabulary", but for us it's not sensible, becasue sooner
   * or later some spam message will make use of that character in
   * some combination with another character. When that happens, we're
   * in trouble.
   *
   * The hack is therefore: we pretend we saw a DIAMOND-j-DIAMOND
   * double transition, for each j. Equivalently, we simply increment
   * the count by 1 for each such transition.
*/
void make_dirichlet_digrams() {
  alphabet_size_t i, j;
  token_count_t k;
  float G[ASIZE];
  float H[ASIZE];
  float V[ASIZE];
  float F[ASIZE];
  double prev_alpha;
  double K, t;

  /* initialize */
  dirichlet.alpha = 1.0;
  for(i = AMIN; i < ASIZE; i++) {
    dirichlet.u[i] = 1.0/((double)(ASIZE - AMIN));
    /* hack see above */
    learner.dig[i][DIAMOND]++;
    learner.dig[DIAMOND][i]++;
  }

  /* precompute useful constants */
  for(j = AMIN; j < ASIZE; j++) {
    G[j] = 0.0;
    H[j] = 0.0;
    V[j] = 0.0;
    F[j] = 0.0;
    for(i = AMIN; i < ASIZE; i++) {
      for(k = AMIN; k < (token_count_t)learner.dig[i][j]; k++) {
	G[j] += 1.0/k;
	H[j] += 1.0/((double)k*k);
      }
      if( learner.dig[i][j] > 0 ) { 
	V[j]++; 
	F[j] += learner.dig[i][j];
      }
    }
    if(0 && u_options & (1<<U_OPTION_DEBUG) ) { 
      fprintf(stdout, 
	      "symbol %d has G = %g H = %g V = %g F = %g\n", 
	      j, G[j], H[j], V[j], F[j]);
    }
  }

  /* now optimize the dirichlet parameters */
  do {
    prev_alpha = dirichlet.alpha;
    K = 0.0;
    for(i = AMIN; i < ASIZE; i++) {
      K += log((F[i] + prev_alpha)/prev_alpha) +
	(0.5) * F[i]/(prev_alpha * (F[i] + prev_alpha));
    }
    dirichlet.alpha = 0.0;
    for(j = AMIN; j < ASIZE; j++) {
      t = K - G[j];
      dirichlet.u[j] = 2 * V[j] /(t + sqrt(t * t + 4 * H[j] * V[j]));
      dirichlet.alpha += dirichlet.u[j];
    }
    if( u_options & (1<<U_OPTION_VERBOSE) ) {
      fprintf(stdout, 
	      "dirichlet alpha %g --> %g\n", 
	      prev_alpha, dirichlet.alpha);
    }
  } while( fabs(prev_alpha - dirichlet.alpha) > qtol_alpha );
  
  if(0 && u_options & (1<<U_OPTION_DEBUG) ) {
    for(j = AMIN; j < ASIZE; j++) {
      fprintf(stdout, 
	      "dirichlet u[%d] =  %g\n", 
	      j, dirichlet.u[j]);
    }
  }

  /* now make the probabilities */

  for(i = AMIN; i < ASIZE; i++) {
    t = 0.0;
    for(j = AMIN; j < ASIZE; j++) {
      t += learner.dig[i][j];
    }
    for(j = AMIN; j < ASIZE; j++) {
      /* note: simulate the effect of digitizing the digrams */
      learner.dig[i][j] = 
	UNPACK_DIGRAMS(PACK_DIGRAMS(log(((weight_t)learner.dig[i][j] + 
					 dirichlet.u[j]) / 
					(t + dirichlet.alpha)))); 
      if(0 && u_options & (1<<U_OPTION_DEBUG) ) {
	fprintf(stdout,
		"learner.dig[%d][%d] = %f\n", 
		i, j, learner.dig[i][j]);
      }
    }
  }
}

void make_uniform_digrams() {
  alphabet_size_t i, j;
  for(i = AMIN; i < ASIZE; i++) {
    for(j = AMIN; j < ASIZE; j++) {
      /* note: simulate the effect of digitizing the digrams */
      learner.dig[i][j] = UNPACK_DIGRAMS(PACK_DIGRAMS(-log(ASIZE - AMIN)));
      if(0 && u_options & (1<<U_OPTION_DEBUG) ) {
	fprintf(stdout,
		"learner.dig[%d][%d] = %f\n", 
		i, j, learner.dig[i][j]);
      }
    }
  }
}

/* use 0.0 for small datasets, 1.0 for huge datasets. 
 * 0.6 works well generally. I don't understand this :-( 
 */
void transpose_digrams() {
  register alphabet_size_t i, j;
  register weight_t t;

  /* we skip transitions involving DIAMOND, TOKENSEP */
  /* code below uses fact that DIAMOND == 0x01 */
  /* code below uses fact that TOKENSEP == 0x02 */
  for(i = AMIN; i < ASIZE; i++) {
    for(j = AMIN; j < i; j++) {
      t = (learner.dig[i][j] + learner.dig[j][i])/2.0;
      learner.dig[j][i] =  0.6 * learner.dig[j][i] + 0.4 * t;
      learner.dig[i][j] =  0.6 * learner.dig[i][j] + 0.4 * t; 
    }
  }
}

void recompute_ed(weight_t *plogzon, weight_t *pdiv, 
		  int i, weight_t lam[ASIZE], weight_t Xi) {
  register alphabet_size_t j;
  weight_t logA = log((weight_t)(ASIZE - AMIN));
  weight_t logzon, div;
  weight_t maxlogz, maxlogz2, tmp;

  maxlogz = 0.0;
  maxlogz2 = log(0.0);
  for(j = AMIN; j < ASIZE; j++) {
    if( lam[j] > 0.0 ) {
      tmp = lam[j] + log(1.0 - exp(-lam[j])) - logA;
      if( maxlogz < tmp ) {
	maxlogz2 = maxlogz;
	maxlogz = tmp;
      }
    }
  }


  logzon = exp(0.0 - maxlogz);
  div = 0.0;
  for(j = AMIN; j < ASIZE; j++) {
    if( lam[j] > 0.0 ) {
      tmp = -logA - maxlogz;
      logzon += (exp( lam[j] + tmp) - exp(tmp));
      div += lam[j] * learner.dig[i][j];
    }
  }


  tmp = maxlogz + log(logzon);

  if( isnan(tmp) ) {
    fprintf(stderr, "error: sorry, digramic partition function went kaboom.\n");
    exit(0);
  }

  logzon = tmp;
  div = div/Xi - logzon;

  *plogzon = logzon;
  *pdiv = div;
}

void make_entropic_digrams() {
  weight_t lam[ASIZE];
  weight_t lam_delta, old_lam, logt, maxlogt;
  weight_t Xi, logXi, logzon, div, old_logzon, old_div;
  weight_t logA = log((weight_t)(ASIZE - AMIN));
  register alphabet_size_t i, j;
  int itcount;

  for(i = AMIN; i < ASIZE; i++) {

    /* initializations */
    Xi = 0.0;
    for(j = AMIN; j < ASIZE; j++) {
      Xi += learner.dig[i][j];
      lam[j] = 0.0;
    }
    logXi = log(Xi);

    recompute_ed(&logzon, &div, i, lam, Xi);

    /* now iterate - this is like minimize_learner_divergence() */
    itcount = 0;
    do {
      itcount++;

      old_logzon = logzon;
      old_div = div;

      lam_delta = 0.0;
      for(j = AMIN; j < ASIZE; j++) {
	if( learner.dig[i][j] > 0.0 ) {

	  old_lam = lam[j];
	  lam[j] = log(learner.dig[i][j]) - logXi + logA + logzon;

	  if( isnan(lam[j]) ) {

	    lam[j] = old_lam;
	    recompute_ed(&logzon, &div, i, lam, Xi);

	  } else {

	    if( lam[j] > (old_lam + MAX_LAMBDA_JUMP) ) {
	      lam[j] = (old_lam + MAX_LAMBDA_JUMP);
	    } else if( lam[j] < (old_lam - MAX_LAMBDA_JUMP) ) {
	      lam[j] = (old_lam - MAX_LAMBDA_JUMP);
	    }

	    /* don't want negative weights */
	    if( lam[j] < 0.0 ) { lam[j] = 0.0; }

	    if( lam_delta < fabs(lam[j] - old_lam) ) {
	      lam_delta = fabs(lam[j] - old_lam);
	    }

	  }
	}
      }


      recompute_ed(&logzon, &div, i, lam, Xi);

      if( u_options & (1<<U_OPTION_VERBOSE) ) {
	fprintf(stdout, "digramic (%d) entropy change %" FMT_printf_score_t \
		" --> %" FMT_printf_score_t " (%10f, %10f)\n", i,
		old_div, div, lam_delta, fabs(logzon - old_logzon));
      }

    
    } while( ((fabs(div - old_div) > 0.001) || (lam_delta > 0.001) ||
	      (fabs(logzon - old_logzon) > 0.001)) && (itcount < 500) );


    /* now make the probabilities */

    /* transitions should already sum to 1, but better safe than sorry */
    maxlogt = log(0.0);
    for(j = AMIN; j < ASIZE; j++) {
      if( maxlogt < lam[j] ) {
	maxlogt = lam[j];
      }
    }
    maxlogt += -logA -logzon;
    logt = 0.0;
    for(j = AMIN; j < ASIZE; j++) {
      logt += exp(lam[j] - logA - logzon - maxlogt);
    }
    logt = maxlogt + log(logt);

    for(j = AMIN; j < ASIZE; j++) {
      /* note: simulate the effect of digitizing the digrams */
      learner.dig[i][j] = 
	UNPACK_DIGRAMS(PACK_DIGRAMS(lam[j] - logA - logzon - logt));

      if(0 && u_options & (1<<U_OPTION_DEBUG) ) {
	fprintf(stdout,
		"learner.dig[%d][%d] = %f\n", 
		i, j, learner.dig[i][j]);
      }
    }
  }
}


void interpolate_digrams(weight_t N, weight_t count[], weight_t p[]) {
  weight_t lambda = 0.5;
  weight_t Clam, Dlam, Zlam;
  alphabet_size_t i;
  int q = 0;

  for(q = 0; q < 100; q++) {
    Zlam = Dlam = Clam = 0.0;
    for(i = AMIN; i < ASIZE; i++) {
      Clam += log(p[i]) * (count[i]/N);
      Zlam += pow(p[i], lambda);
      Dlam += log(p[i]) * pow(p[i], lambda);
    }
    lambda += 0.1 * (Clam - Dlam/Zlam);
/*     if( lambda > 1.0 ) { lambda = 1.0; } */
/*     else if( lambda < 0.0 ) { lambda = 0.0; } */
/*     printf("Clam = %g Dlam = %g, Zlam = %g, lamdba = %g\n", Clam, Dlam, Zlam, lambda); */
  } 
  printf("interpolating %g\n", lambda);

}


weight_t calc_learner_digramic_excursion(char *tok) {
  register alphabet_size_t p, q;
  register weight_t t = 0.0;

  /* now update digram frequency counts */
  p = (unsigned char)*tok++;
  /* finally update character frequencies */
  while( *tok != EOTOKEN ) {
    q = (unsigned char)*tok;
    t += learner.dig[p][q];
    p = q;
    tok++;
  }
  return t;
}

/* calculates the rth-order divergence but needs normalizing constant
   note: this isn't the full divergence from digref, just the bits
   needed for the r-th optimization - fwd indicates the traversal
   direction in the hash (alternate between them to reduce numerical
   errors */
score_t learner_divergence(score_t logzonr, score_t Xi,
			   token_order_t r, bool_t fwd) {
  register l_item *i, *e;
  register token_count_t c = 0;
  score_t t = 0.0;

  e = fwd ? (learner.hash + learner.max_tokens) : learner.hash - 1;
  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {

      t += i->lam * (score_t)i->count;

      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

  return -logzonr + t/Xi;
}

/* calculates the normalizing constant - fwd indicates the traversal
   direction in the hash (alternate between them to reduce numerical
   errors */
score_t learner_logZ(token_order_t r, score_t log_unchanging_part, bool_t fwd) {
  register l_item *i, *e;
  register token_count_t c = 0;

  score_t maxlogz, maxlogz2, tmp;
  score_t t =0.0;
  score_t R = (score_t)r;

  e = fwd ? (learner.hash + learner.max_tokens) : learner.hash - 1;
  maxlogz = log_unchanging_part;
  maxlogz2 = log(0.0);
  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {

      if( NOTNULL(i->lam) ) {
	tmp = R * (i->lam + UNPACK_LWEIGHTS(i->ltrms)) + 
	  log(1.0 - exp(-R * i->lam)) + UNPACK_RWEIGHTS(i->dref);

	if( maxlogz < tmp ) {
	  maxlogz2 = maxlogz;
	  maxlogz = tmp;
	}
      }
      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

  t =  exp(log_unchanging_part - maxlogz);
  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {

      tmp = R * UNPACK_LWEIGHTS(i->ltrms) + UNPACK_RWEIGHTS(i->dref) - maxlogz;
      t += (exp(R * i->lam + tmp) - exp(tmp)); 

      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

  tmp =  (maxlogz + log(t))/R;
/*   printf("t =%f maxlogz = %f logZ/R = %f\n", t, maxlogz, tmp); */

  /* investigate this someday */
  if( isnan(tmp) ) {
    fprintf(stderr, "error: sorry, partition function went kaboom.\n");
    exit(0);
  }

  return tmp;
}

/* fills the hash with partial calculations */
void fill_ref_vars(l_item *k, char *tok, token_order_t r) {
  hash_value_t id;
  char *t;
  l_item *l;

  /* weight of the r-th order excursion */
  k->dref = PACK_RWEIGHTS(calc_learner_digramic_excursion(tok));

  /* for each prefix of tok, add its weight */
  k->ltrms = PACK_LWEIGHTS(0.0);
  if( r > 1 ) {
    for( t = tok + strlen(tok) - EXTRA_CLASS_LEN - 1; t >= tok; t-- ) {
      if( *(t - 1) == DIAMOND ) {
	*t = '\0';
	id = (hash_value_t)hash((unsigned char *)tok, t - tok, 0);
	l = find_in_learner(id); 
	if( l ) {
	  k->ltrms += PACK_LWEIGHTS(l->lam);
	}
      }
    }
  }

}

/* fills hash with partial calculations and returns the
   log unchanging part of the normalizing constant */
score_t recalculate_reference_measure(token_order_t r, score_t *kappa) {
  hash_value_t id;
  char buf[BUFSIZ+1];
  char tok[MAX_TOKEN_LEN];
  size_t n = 0;

  char *p, *q;
  l_item *k, *i;

  score_t tmp, lunch;
  score_t R = (score_t)r;
  score_t max = 0.0;
  score_t mykappa = 0.0;

  /* for r == 1, the partial_z is not modified, but
     we still have to go through the tokens and prebuild k-dref */

  /* now we calculate the logarithmic word weight
     from the digram model, for each token in the hash */
  rewind(learner.tmp);
  q = tok;
  while( (n >= 0) && !feof(learner.tmp) ) {
    n = fread(buf, 1, BUFSIZ, learner.tmp);
    if( n > 0 ) {
      p = buf;
      p[n] = '\0';
      while( *p ) {
	if( *p != TOKENSEP) {
	  *q++ = *p; /* copy into tok */
	} else { /* interword space */ 
	  *q = 0; /* append NUL to tok */
	  /* now write weight in hash */
	  id = (hash_value_t)hash((unsigned char *)tok, strlen(tok), 0);
	  k = find_in_learner(id); 
	  if( k ) {
	    fill_ref_vars(k, tok, r);

	    if( k->typ.order == r ) {
	      mykappa += exp(UNPACK_RWEIGHTS(k->dref));
	    } else if(k->typ.order < r) {
	      if( k->lam > 0.0 ) {
		tmp = UNPACK_RWEIGHTS(k->dref) + 
		  R * (k->lam + UNPACK_LWEIGHTS(k->ltrms));
		if( max < tmp ) {
		  max = tmp;
		}
	      }
	    }
	  }
	  q = tok; /* reset q */
	}
	p++;
      }
    }
  }
  /* finish last word */
  *q = 0; /* append NUL to tok */
  /* now write weight in hash */
  id = (hash_value_t)hash((unsigned char *)tok, strlen(tok), 0);
  k = find_in_learner(id); /* guaranteed to be found */
  if( k ) {
    fill_ref_vars(k, tok, r);	

    if( k->typ.order == r ) {
      mykappa += exp(UNPACK_RWEIGHTS(k->dref));
    } else if(k->typ.order < r) {
      if( k->lam > 0.0 ) {
	tmp = UNPACK_RWEIGHTS(k->dref) + 
	  R * (k->lam + UNPACK_LWEIGHTS(k->ltrms));
	if( max < tmp ) {
	  max = tmp;
	}
      }
    }
  }
  /* done */

  lunch = 0.0;
  k = learner.hash + learner.max_tokens;
  for(i = learner.hash; i != k; i++) {
    if( FILLEDP(i) ) {
      if( i->typ.order < r ) {
	if( NOTNULL(i->lam) ) {
	  tmp = -max + R * UNPACK_LWEIGHTS(i->ltrms) + UNPACK_RWEIGHTS(i->dref);
	  lunch += (exp(R * i->lam + tmp) - exp(tmp));
	}
      }
    }
  }
  lunch = max + lunch;

  *kappa = mykappa;

  return lunch;
}


/* just for debugging */
void print_score(token_order_t r) {
  hash_count_t i;
  double logprob = 0.0;
  double lpapprox = 0.0;
  double norm;

  for(i = 0; i < learner.max_tokens; i++) {
    if( FILLEDP(&learner.hash[i]) &&
	(learner.hash[i].typ.order <= r) ) {
      logprob += learner.hash[i].count * (learner.hash[i].lam);
      if( learner.hash[i].typ.order == 1 ) {
	logprob += learner.hash[i].count * 
	  UNPACK_RWEIGHTS(learner.hash[i].dref)/((weight_t)r);
      }
    }
    if( FILLEDP(&learner.hash[i]) &&
	(learner.hash[i].typ.order == r) ) {
      lpapprox += learner.hash[i].count * 
	((learner.hash[i].lam) + 
	 UNPACK_LWEIGHTS(learner.hash[i].ltrms) + 
	 UNPACK_RWEIGHTS(learner.hash[i].dref)/((weight_t)r));
    }

  }
  norm = learner.fixed_order_token_count[r] * learner.logZ;

  fprintf(stdout,
	  "*** logprob = %" FMT_printf_score_t " * %d (r = %d, logZ = %" FMT_printf_score_t ")\n", 
	 ((score_t)(logprob / learner.fixed_order_token_count[r])), 
	 learner.fixed_order_token_count[r], 
	 r, learner.logZ);
  fprintf(stdout,
	  "*** lpapprox = %" FMT_printf_score_t " * %d (r = %d, logZ = %" FMT_printf_score_t ")\n", 
	 ((score_t)(lpapprox / learner.fixed_order_token_count[r])), 
	 learner.fixed_order_token_count[r], 
	 r, learner.logZ);
}


/* experimental: this sounded like a good idea, but isn't ?? */
void theta_rescale(token_order_t r, score_t logupz, score_t Xi, bool_t fwd) {
  register l_item *i, *e;
  register token_count_t c = 0;

  score_t tmp;
  score_t theta =0.0;
  score_t R = (score_t)r;
  score_t mu;
  score_t sum1, sum2;
  score_t msum1, msum2;
  token_count_t s = 0;

  e = fwd ? (learner.hash + learner.max_tokens) : learner.hash - 1;

  msum1 = msum2 = log(0.0);
  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {
      if( NOTNULL(i->lam) ) {
	s += i->count;
	tmp = R * UNPACK_LWEIGHTS(i->ltrms) + UNPACK_RWEIGHTS(i->dref);
	if( tmp > msum1 ) {
	  msum1 = tmp;
	}
	tmp += R * i->lam;
	if( tmp > msum2 ) {
	  msum2 = tmp;
	}
      }
      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }  

  mu = ((score_t)s)/Xi;
  s = log(mu) - log(1 - mu) + logupz; 
  if( s > msum1 ) {
    msum1 = s;
  }

  sum1 = exp(s - msum1);
  sum2 = 0.0;
  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {
      if( NOTNULL(i->lam) ) {
	tmp = R * UNPACK_LWEIGHTS(i->ltrms) + UNPACK_RWEIGHTS(i->dref);
	sum1 += exp(tmp - msum1);
	tmp += R * i->lam;
	sum2 += exp(tmp - msum2);
      }
      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

  theta = ((msum1 + log(sum1)) - (msum2 + log(sum2)))/R;
  if( isnan(theta) ) {
    return; /* ignore */
  }
  if( u_options & (1<<U_OPTION_VERBOSE) ) {
    fprintf(stdout, "theta = %f\n", theta);
  }

  for(i = fwd ? learner.hash : learner.hash + learner.max_tokens - 1; 
      i != e; fwd ? i++ : i--) {
    if( FILLEDP(i) && (i->typ.order == r) ) {
      if( NOTNULL(i->lam) ) {
	i->lam += theta;
      }
      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

}

/* minimizes the divergence by solving for lambda one 
   component at a time.  */
void minimize_learner_divergence() {
  l_item *i, *e;
  token_order_t r;
  token_count_t lzero, c = 0;
  int itcount, mcount;
  score_t d, dd, b;
  score_t lam_delta, old_lam;
  score_t logzonr, old_logzonr;
  score_t logupz, div_extra_bits, kappa, thresh;
  score_t R, Xi, logXi;
  score_t mp_logz;
  bool_t fwd = 1;

  if( u_options & (1<<U_OPTION_VERBOSE) ) {
    fprintf(stdout, "now maximizing model entropy\n");
  }

  learner.logZ = 0.0;
  learner.divergence = 0.0;
  mp_logz = 0.0;

  /* disable multipass if we only have one order to play with */
  if( (m_options & (1<<M_OPTION_MULTINOMIAL)) || 
      (learner.max_order == 1) ) {
    qtol_multipass = 0;
  }

  e = learner.hash + learner.max_tokens;
  for(mcount = 0; mcount < (qtol_multipass ? 50 : 1); mcount++) {
    for(r = 1; 
	r <= ((m_options & (1<<M_OPTION_MULTINOMIAL)) ? 1 : learner.max_order); 
	r++) {

      /* here we precalculate various bits and pieces
	 which aren't going to change during this iteration */
      logupz = recalculate_reference_measure(r, &kappa);

      R = (score_t)r;
      Xi = (score_t)learner.fixed_order_token_count[r];
      logXi = log(Xi);

      if( r == 1 ) {
/*       Xi /= kappa; */
	div_extra_bits = 0.0;
      } else {
      /* calculate extra bits for divergence score display */
	b = 0.0;

	for(i = learner.hash; i != e; i++) {
	  if( FILLEDP(i) && (i->typ.order < r) ) {
	    b += i->lam * (score_t)i->count;
	  }
	}
	div_extra_bits = b/Xi;
      }

      if( m_options & (1<<M_OPTION_REFMODEL) ) {
	break;
      }

      if( u_options & (1<<U_OPTION_VERBOSE) ) {
	fprintf(stdout, "* optimizing order %d weights (%i tokens, %i unique, kappa = %f)\n", 
		r, learner.fixed_order_token_count[r],
		learner.fixed_order_unique_token_count[r], kappa);
      }

      logzonr = learner_logZ(r, logupz, fwd);
      dd = learner_divergence(logzonr, Xi, r, fwd);
      fwd = 1 - fwd;

      itcount = 0;
      thresh = 0.0;
      do {
	itcount++;
	lzero= 0;

	d = dd;     /* save old divergence */
	old_logzonr = logzonr;

	lam_delta = 0.0;

	c = 0;
	for(i = learner.hash; i != e; i++) {
	  if( FILLEDP(i) && (i->typ.order == r) 
/* 	    && (rand() > RAND_MAX/2)  */
	      ) { 

	    old_lam = i->lam;


	    /* "iterative scaling" lower bound */
	    i->lam = (log((score_t)i->count) - logXi -  
		      UNPACK_RWEIGHTS(i->dref))/R + logzonr -
	      UNPACK_LWEIGHTS(i->ltrms);

	    if( isnan(i->lam) ) {
	      /* precision problem, just ignore */
	      i->lam = old_lam;
	      logzonr = learner_logZ(r, logupz, fwd);
	      thresh = 0.0;
	    } else {

	      /* this code shouldn't be necessary, but is crucial */
	      if( i->lam > (old_lam + MAX_LAMBDA_JUMP) ) {
		i->lam = (old_lam + MAX_LAMBDA_JUMP);
	      } else if( i->lam < (old_lam - MAX_LAMBDA_JUMP) ) {
		i->lam = (old_lam - MAX_LAMBDA_JUMP);
	      }

	      /* don't want negative weights */
	      if( i->lam < 0.0 ) { i->lam = 0.0; lzero++; }

	      if( lam_delta < fabs(i->lam - old_lam) ) {
		lam_delta = fabs(i->lam - old_lam);
	      }

	    }

	    if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	      c = 0;
	      break;
	    }
	  }
	}

/* 	theta_rescale(r, logupz, Xi, fwd); */

	/* update values */
	logzonr = learner_logZ(r, logupz, fwd);
	dd = learner_divergence(logzonr, Xi, r, fwd);
	fwd = 1 - fwd;

	if( u_options & (1<<U_OPTION_VERBOSE) ) {
/* 	fprintf(stdout, "lzero = %ld\n", lzero); */
	  fprintf(stdout, "entropy change %" FMT_printf_score_t \
		  " --> %" FMT_printf_score_t " (%10f, %10f)\n", 
		  d + div_extra_bits, 
		  dd + div_extra_bits,
		  lam_delta, fabs(logzonr - old_logzonr));
	}

	process_pending_signal(NULL);
      
      } while( ((fabs(d - dd) > qtol_div) || (lam_delta > qtol_lam) ||
		(fabs(logzonr - old_logzonr) > qtol_logz)) && (itcount < 50) );

      learner.logZ = logzonr;
      learner.divergence = dd + div_extra_bits;
    }
    /* for multipass, we wait until logZ stabilizes */
    if( fabs(1.0 - mp_logz/learner.logZ) < 0.01 ) {
      break;
    }
    mp_logz = learner.logZ;
  }

}

/* dumps readable model weights to the output */
void dump_model(FILE *out, FILE *in) {

  hash_value_t id;
  char buf[BUFSIZ+1];
  char tok[MAX_TOKEN_LEN];
  char smb[MAX_SUBMATCH+1];
  token_order_t s;
  regex_count_t c;
  char *p, *q;
  l_item *k;
  size_t n = 0;

  /* preamble - this is copied from save_learner */

  fprintf(out, MAGIC1, learner.filename, 
	  (m_options & (1<<M_OPTION_REFMODEL)) ? "(ref)" : "");
  fprintf(out, MAGIC2_o, learner.divergence, learner.logZ, learner.max_order,
	  (m_options & (1<<M_OPTION_MULTINOMIAL)) ? "multinomial" : "hierarchical" );
  fprintf(out, MAGIC3, 
	  (short int)learner.max_hash_bits, 
	  (long int)learner.full_token_count, 
	  (long int)learner.unique_token_count,
	  (long int)learner.num_docs);

  /* print out any regexes we might need */
  for(c = 0; c < regex_count; c++) {
    /* write the bitmap */
    for(p = smb, s = 1; s <= MAX_SUBMATCH; s++) {
      if( re[c].submatches & (1<<s) ) {
	*p++ = s + '0';
      }
    }
    *p = '\0';

#if defined HAVE_LIBBOOST_REGEX
    /* does this work? */
    fprintf(out, MAGIC5_wo, re[c].string, smb);
#else
    fprintf(out, MAGIC5_o, re[c].string, smb);
#endif
  }

  /* print options */
  fprintf(out, MAGIC4_o, m_options, print_model_options(m_options, buf));

  fprintf(out, MAGIC6); 

  fprintf(out, "# lambda | dig_ref | count | token\n");


  /* now go through hash printing values */
  rewind(learner.tmp);
  q = tok;
  while( (n >= 0) && !feof(learner.tmp) ) {
    n = fread(buf, 1, BUFSIZ, learner.tmp);
    if( n > 0 ) {
      p = buf;
      p[n] = '\0';
      while( *p ) {
	if( *p != TOKENSEP) {
	  *q++ = *p; /* copy into tok */
	} else { /* interword space */ 
	  *q = 0; /* append NUL to tok */
	  /* now write weight in hash */
	  id = (hash_value_t)hash((unsigned char *)tok, strlen(tok), 0);
	  k = find_in_learner(id); /* guaranteed to be found */
	  fprintf(out, "%9.3f %9.3f %7d %s\n", 
		  (weight_t)k->lam, UNPACK_RWEIGHTS(k->dref), k->count, tok);
	  q = tok; /* reset q */
	}
	p++;
      }
    }
  }
  /* finish last word */
  *q = 0; /* append NUL to tok */
  /* now write weight in hash */
  id = (hash_value_t)hash((unsigned char *)tok, strlen(tok), 0);
  k = find_in_learner(id); /* guaranteed to be found */
  fprintf(out, "%9.3f %9.3f %7d %s\n", 
	  (weight_t)k->lam, UNPACK_RWEIGHTS(k->dref), k->count, tok);

}

/* This is a quick hack to 
   1) speed up the optimization and 
   2) stabilise the estimated category model over a sequence of learning runs.

   Normally, if the dataset is nearly the same, then the optimal lambda weights 
   should be fairly similar (This is not strictly true if the digramic
   measure varies too much).  

   By reading the old category and presetting the lambdas from it, we typically
   get close enough to the true weights that one or two optimization iterations
   suffice. Moreover in that case, most weights should stay close to their
   "previous" values in the category, so the user doesn't think it's weird.

   A side effect is that there is entropy creep when relearing the same
   dataset over and over. Unless there's a bug I haven't noticed, this occurs
   because the entropy maximum we compute is affected by numerical precision
   and feature ordering within the hash. Each time we relearn the same dataset,
   we start off with a perturbation from the previous best weights, and obtain 
   a small improvement.
   
 */
void learner_prefill_lamdas() {
  hash_count_t c;
  c_item *i, *e;
  l_item *k;
  Category dummycat;

  dummycat.fullfilename = strdup(learner.filename);
  if( load_category(&dummycat) ) {
    if( dummycat.m_options & (1<<M_OPTION_WARNING_BAD) ) {
      if( u_options & (1<<U_OPTION_VERBOSE) ) {
	fprintf(stderr, "warning: old category file %s may have bad data\n",
		dummycat.fullfilename);
      }
    } else if( (dummycat.m_options == m_options) &&
	       (dummycat.retype == learner.retype) &&
	       (fabs((dummycat.model_unique_token_count/
		      (double)learner.unique_token_count) - 1.0) < 0.15) ) {
      c = dummycat.model_unique_token_count;
      if( u_options & (1<<U_OPTION_VERBOSE) ) {
	fprintf(stdout, "preloading %ld token weights\n", (long int)c);
      }
    
#if defined HAVE_MADVISE
      madvise(dummycat.hash, sizeof(c_item) * dummycat.max_tokens, MADV_SEQUENTIAL);
#endif
      e = dummycat.hash + dummycat.max_tokens;
      for(i = dummycat.hash; i != e; i++) {
	if( FILLEDP(i) ) {
	  k = find_in_learner(i->id);
	  if( k ) {
	    k->lam = UNPACK_LAMBDA(i->lam);
	  }
	  if( --c <= 0 ) {
	    break;
	  }
	}
      }

      /* tweak default quality settings */
      if( !quality ) {
	qtol_div = 0.01;
	qtol_lam = 0.01;
	qtol_logz = 0.05;
      }

    }
    free_category(&dummycat);
  } else {
    if( u_options & (1<<U_OPTION_VERBOSE) ) {
      fprintf(stderr, "warning: cannot open file for reading %s\n", 
	      dummycat.fullfilename);
    }
  }
}


void optimize_learner_and_save() {

  hash_count_t i;
  token_order_t c;

  if(100 * learner.unique_token_count >= HASH_FULL * learner.max_tokens) { 
    fprintf(stderr, 
	    "warning: table full, some tokens ignored - "
	    "try with option -h %i\n",
	    learner.max_hash_bits + 1);
  } else if( learner.unique_token_count <= 0 ) {
    fprintf(stderr, 
	    "warning: no tokens matched - have I learned nothing?\n");
    exit(1);
  }

  if( skewed_constraints_warning ) {
    fprintf(stderr,
	    "warning: ran out of integers (too much data) constraints will be skewed.\n");
    m_options |= (1<<M_OPTION_WARNING_BAD);
  }

  if( u_options & (1<<U_OPTION_VERBOSE) ) {
    fprintf(stdout, 
	    "picked up %i (%i distinct) tokens\n", 
	    learner.full_token_count, learner.unique_token_count);
    fprintf(stdout, 
	    "calculating reference word weights\n");
  }

  if( *online ) {
    write_online_learner_struct(online);
  }

  /* transposition smoothing */
/*   transpose_digrams(); */

  if( u_options & (1<<U_OPTION_LAPLACE) ) {
    make_uniform_digrams();
  } else if( u_options & (1<<U_OPTION_DIRICHLET) ) {
    make_dirichlet_digrams();
  } else if( u_options & (1<<U_OPTION_JAYNES) ) {
    make_entropic_digrams();
  } else {
    make_uniform_digrams();
  }

  if( learner.fixed_order_token_count[1] == 0 ) {
    /* it's a higher order model but there are no first
       order tokens! We can't handle that! Go through the 
       hash converting everything to first order.
       This will result in incorrect calculations, unless
       the higher order tokens don't overlap. Suitable only for
       geniuses and fools. */
    fprintf(stderr, 
	    "\n"
	    "warning: you have not defined any unigrams, so in this model\n"
	    "         features will be treated independently, which is quite\n"
	    "         likely incorrect. I hope you know what you're doing,\n"
	    "         because I don't!\n\n");

    m_options |= (1<<M_OPTION_MULTINOMIAL);
    for(c = 2; c <= learner.max_order; c++) {
      learner.fixed_order_token_count[1] += learner.fixed_order_token_count[c];
      learner.fixed_order_unique_token_count[1] += 
	learner.fixed_order_unique_token_count[c];
    }
    for(i = 0; i < learner.max_tokens; i++) {
      if( FILLEDP(&learner.hash[i]) ) {
	learner.hash[i].typ.order = 1;
      }
    }

  }


  /* if the category already exists, we read its lambda values.
     this should speed up the minimization slightly */
  if( !(u_options & (1<<U_OPTION_ZEROLEARN)) ) {
    learner_prefill_lamdas();
  }

  minimize_learner_divergence();

  if( u_options & (1<<U_OPTION_DUMP) ) {
    rewind(learner.tmp);
    dump_model(stdout, learner.tmp);
  }
  fclose(learner.tmp);
  /* we don't free learner.tmpiobuf, as we'll be exiting soon anyway */
#if defined HAVE_POSIX_MEMALIGN
/*   if( learner.tmpiobuf ) { free(learner.tmpiobuf); } */
#endif
  /* now save the model to a file */
  save_learner();
}

/***********************************************************
 * MULTIBYTE FILE HANDLING FUNCTIONS                       *
 * this is suitable for any locale whose character set     *
 * encoding doesn't include NUL bytes inside characters    *
 ***********************************************************/

/* this code executed before processing each line of input.
   - handles indents and appends via -Aa switches  */
char *handle_indents_and_appends(char *textbuf) {

  char *pptextbuf = textbuf; /* default */

  if( u_options & (1<<U_OPTION_INDENTED) ) {
    if( textbuf[0] == ' ' ) {
      pptextbuf = textbuf + 1; /* processing should ignore indent */
    } else {
      if( u_options & (1<<U_OPTION_APPEND) ) {
	fprintf(stdout, "%s", textbuf);
      }
      pptextbuf = NULL; /* no further processing */
    }
  }

  /* if appending, print the lines as they 
     come in before they are processed */
  if( u_options & (1<<U_OPTION_APPEND) ) {
    fprintf(stdout, " %s", pptextbuf);
  }

  return pptextbuf;
}

/***********************************************************
 * WIDE CHARACTER FILE HANDLING FUNCTIONS                  *
 * this is needed for any locale whose character set       *
 * encoding can include NUL bytes inside characters        *
 ***********************************************************/


/***********************************************************
 * MAIN FUNCTIONS                                          *
 ***********************************************************/


int main(int argc, char **argv) {

  category_count_t c;
  FILE *input;
  struct stat statinfo;
  signed char op;

  void (*preprocess_fun)(void) = NULL;
  void (*word_fun)(char *, token_type_t, regex_count_t) = NULL;
  char *(*pre_line_fun)(char *) = NULL;
  void (*post_line_fun)(char *) = NULL;
  void (*postprocess_fun)(void) = NULL;

  learner.filename = NULL;
  learner.retype = 0;

  /* set up internationalization */
  if( !setlocale(LC_ALL, "") ) {
    fprintf(stderr, 
	    "warning: could not set locale, internationalization disabled\n");
  } else {
    if( u_options & (1<<U_OPTION_VERBOSE) ) {
      fprintf(stderr, 
	      "warning: international locales not supported\n");
    }
  }

#if defined(HAVE_GETPAGESIZE)
  system_pagesize = getpagesize();
#endif
  if( system_pagesize == -1 ) { system_pagesize = BUFSIZ; }

  init_signal_handling();

  /* parse the options */
  while( (op = getopt(argc, argv, "0Aac:Dde:f:G:g:H:h:ijL:l:MNno:q:RrT:Vvw:x:")) > -1 ) {

    switch(op) {
    case '0':
      u_options |= (1<<U_OPTION_ZEROLEARN);
      break;
    case 'A':
      u_options |= (1<<U_OPTION_INDENTED);
      /* fall through */
    case 'a':
      u_options |= (1<<U_OPTION_APPEND);
      break;
    case 'e':
      if( !strcasecmp(optarg, "alnum") ) {
	m_options |= (1<<M_OPTION_CHAR_ALNUM);
      } else if( !strcasecmp(optarg, "alpha") ) {
	m_options |= (1<<M_OPTION_CHAR_ALPHA);
      } else if( !strcasecmp(optarg, "cef") ) {
	m_options |= (1<<M_OPTION_CHAR_CEF);
      } else if( !strcasecmp(optarg, "graph") ) {
	m_options |= (1<<M_OPTION_CHAR_GRAPH);
      } else if( !strcasecmp(optarg, "adp") ) {
	m_options |= (1<<M_OPTION_CHAR_ADP);
      } else {
	fprintf(stderr, "warning: unrecognized option \"%s\", ignoring.\n", 
		optarg);
      }
      break;
    case 'D':
      u_options |= (1<<U_OPTION_DEBUG);
      break;
    case 'M':
      m_options |= (1<<M_OPTION_MULTINOMIAL);
      break;
    case 'N':
      u_options |= (1<<U_OPTION_POSTERIOR);
      break;
    case 'R':
      if( cat_count >= MAX_CAT ) {
	fprintf(stderr, 
		"warning: maximum reached, random text category omitted\n");
      } else if( u_options & (1<<U_OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -R together\n");
	usage(argv);
	exit(0);
      } else {
	u_options |= (1<<U_OPTION_CLASSIFY);

	cat[cat_count].fullfilename = "random";

	init_category(&cat[cat_count]);
	init_purely_random_text_category(&cat[cat_count]);

	cat_count++;
      }
      break;
    case 'T':
      if( !strncasecmp(optarg, "text", 4) ) {
	/* this is the default */
	/* below, make sure the comparisons are in the correct order */
      } else if( !strcasecmp(optarg, "email:noplain") ) {
	m_options |= (1<<M_OPTION_NOPLAIN);
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "email:plain") ) {
	m_options |= (1<<M_OPTION_PLAIN);
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "email:headers") ) {
	m_options |= (1<<M_OPTION_HEADERS);
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "email:atts") ) {
	m_options |= (1<<M_OPTION_ATTACHMENTS);
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "email:xheaders") ) {
	m_options |= (1<<M_OPTION_XHEADERS);
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "email") ) {
	m_options |= (1<<M_OPTION_MBOX_FORMAT);
      } else if( !strcasecmp(optarg, "html:links") ) {
	m_options |= (1<<M_OPTION_SHOW_LINKS);
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "html:alt") ) {
	m_options |= (1<<M_OPTION_SHOW_ALT);
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "html:scripts") ) {
	m_options |= (1<<M_OPTION_SHOW_SCRIPT);
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "html:styles") ) {
	m_options |= (1<<M_OPTION_SHOW_STYLE);
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "html:comments") ) {
	m_options |= (1<<M_OPTION_SHOW_HTML_COMMENTS);
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "html") ) {
	m_options |= (1<<M_OPTION_HTML);
      } else if( !strcasecmp(optarg, "xml") ) {
	m_options |= (1<<M_OPTION_XML);
      } else { /* default */
	fprintf(stderr, 
		"warning: unrecognized option -T %s, ignoring.\n", 
		optarg);
      }
      break;
    case 'V':
      fprintf(stdout, "dbacl version %s\n", VERSION);
#if defined HAVE_LIBBOOST_REGEX
      fprintf(stdout, "Using BOOST wide character modern regexes.\n");
#elif defined __GNUC__
      fprintf(stdout, "Using GNU modern regexes.\n");
#else
      fprintf(stdout, "Using system regexes.\n");
#endif
#if defined DIGITIZE_DIGRAMS
      fprintf(stdout, "Digrams are digitized.\n");
#endif
#if defined DIGITIZE_LAMBDA
      fprintf(stdout, "Lagrangian multipliers are digitized.\n");
#endif
#if defined DIGITIZE_LWEIGHTS
      fprintf(stdout, "Reference weights are digitized.\n");
#endif
      fprintf(stdout, "Feature memory requirements: %d bytes (classifying), %d bytes (learning)\n", 
	      (int)sizeof(c_item), (int)sizeof(l_item));
      fprintf(stdout, "To change these settings, recompile from source.\n");
      exit(1);
      break;
    case 'd':
      u_options |= (1<<U_OPTION_DUMP);
      break;
    case 'h': /* select memory size in powers of 2 */
      default_max_hash_bits = strtod(optarg, NULL);
      if( default_max_hash_bits > MAX_HASH_BITS ) {
	fprintf(stderr, 
		"warning: maximum hash size will be 2^%d\n", 
		MAX_HASH_BITS);
	default_max_hash_bits = MAX_HASH_BITS;
      }
      default_max_tokens = (1<<default_max_hash_bits);
      break;
    case 'H': /* select memory size in powers of 2 */
      default_max_grow_hash_bits = strtod(optarg, NULL);
      if( default_max_grow_hash_bits > MAX_HASH_BITS ) {
	fprintf(stderr, 
		"warning: maximum hash size will be 2^%d\n", 
		MAX_HASH_BITS);
	default_max_grow_hash_bits = MAX_HASH_BITS;
      }
      default_max_grow_tokens = (1<<default_max_grow_hash_bits);
      u_options |= (1<<U_OPTION_GROWHASH);
      break;
    case 'j':
      m_options |= (1<<M_OPTION_CASEN);
      break;
    case 'n':
      u_options |= (1<<U_OPTION_SCORES);
      break;
    case 'c':
      if( cat_count >= MAX_CAT ) {
	fprintf(stderr, 
		"warning: maximum reached, category ignored\n");
      } else if( u_options & (1<<U_OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -c together\n");
	usage(argv);
	exit(0);
      } else {
	u_options |= (1<<U_OPTION_CLASSIFY);

	cat[cat_count].fullfilename = sanitize_path(optarg);

	if( !*optarg ) {
	  fprintf(stderr, "error: category needs a name\n");
	  usage(argv);
	  exit(0);
	}
	if( !load_category(&cat[cat_count]) ) {
	  fprintf(stderr, 
		  "error: could not load category %s\n", 
		  cat[cat_count].fullfilename);
	  exit(0);
	}
	if( sanitize_model_options(&m_options,&cat[cat_count]) ) {
	  ngram_order = (ngram_order < cat[cat_count].max_order) ? 
	    cat[cat_count].max_order : ngram_order;
	  cat_count++;
	}
      }
      break;
    case 'q':
      quality = atoi(optarg);
      switch(quality) {
      case 1:
	qtol_div = 0.01;
	qtol_lam = 0.05;
	qtol_logz = 0.05;
	break;
      case 2:
	qtol_div = 0.01;
	qtol_lam = 0.01;
	qtol_logz = 0.05;
	break;
      case 3:
	qtol_div = 0.01;
	qtol_lam = 0.01;
	qtol_logz = 0.01;
	break; 
      case 4:
	qtol_div = 0.01;
	qtol_lam = 0.01;
	qtol_logz = 0.01;
	qtol_multipass = 1;
	break; 
      default:
	fprintf(stderr, 
		"error: the -q switch needs a number between 1(fastest) and 4(best quality)\n");
	exit(0);
	break;
      }
      break;
    case 'r':
      m_options |= (1<<M_OPTION_REFMODEL);
      break;
    case 'w':
      ngram_order = atoi(optarg);
      if( !*optarg || (ngram_order < 1) || (ngram_order > 9) ) {
	fprintf(stderr, 
		"error: the -w switch needs a number between 1 and 9\n");
	exit(0);
      }
      break;
    case 'x':
      decimation = atoi(optarg);
      if( !*optarg || (decimation < 1) || (decimation > MAX_HASH_BITS) ) {
	fprintf(stderr, 
		"warning: option -x ignored, needs an integer between 1 and %d\n", 
		MAX_HASH_BITS);
      } else {
	u_options |= (1<<U_OPTION_DECIMATE);
      }
      break;
    case 'f':
      if( filter_count >= MAX_CAT ) {
	fprintf(stderr, "warning: maximum reached, filter ignored\n");
      } else if( u_options & (1<<U_OPTION_LEARN) ) {
	fprintf(stderr, "error: cannot use options -l and -f together\n");
	usage(argv);
	exit(0);
      } else if( !*optarg ) {
	fprintf(stderr, "error: filter must be category name or number\n");
	usage(argv);
	exit(0);
      } else {
	u_options |= (1<<U_OPTION_FILTER);
	filter[filter_count] = -1;

	/* see if it's a known category */
	for(c = 0; c < cat_count; c++) {
	  if( !strcmp(cat[c].filename, optarg) ) {
	    filter[filter_count] = c;
	    break;
	  }
	}
	/* if not recognized, see if it's a number */
	if( filter[filter_count] < 0 ) {
	  filter[filter_count] = strtod(optarg, NULL) - 1;
	}
	if( filter[filter_count] < 0 ) { /* still not recognized */
	  fprintf(stderr, 
		  "error: unrecognized category in -f option [%s]\n", 
		  optarg);
	  usage(argv);
	  exit(0);
	}
	filter_count++;
      }
      break;
    case 'v':
      u_options |= (1<<U_OPTION_VERBOSE);
      break;
    case 'L':
      if( *optarg && !strcmp(optarg, "uniform") ) {
	u_options |= (1<<U_OPTION_LAPLACE);
      } else if(*optarg && !strcmp(optarg, "dirichlet") ) {
	u_options |= (1<<U_OPTION_DIRICHLET);
      } else if(*optarg && !strcmp(optarg, "maxent") ) {
	u_options |= (1<<U_OPTION_JAYNES);
      } else {
	fprintf(stderr, 
		"error: -L option needs \"uniform\", \"dirichlet\" or \"maxent\"\n");
	exit(0);
      }
      break;
    case 'l': 
      if( u_options & (1<<U_OPTION_CLASSIFY) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -c together\n");
	usage(argv);
	exit(0);
      } else if( u_options & (1<<U_OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: option -l can only occur once\n");
	exit(0);
      } else if( !*optarg ) {
	fprintf(stderr, "error: category name must not be empty\n");
      } else {
	u_options |= (1<<U_OPTION_LEARN);
	learner.filename = sanitize_path(optarg);
	if( !*learner.filename ) {
	  fprintf(stderr, 
		  "error: category needs a name\n");
	  usage(argv);
	  exit(0);
	}
      }
      break;
    case 'o':
      if( !*optarg ) {
	fprintf(stderr, "error: category name must not be empty\n");
      } else {
	online = sanitize_path(optarg);
      }
      break;
    case 'G':
    case 'g':
      /* if regex can't be compiled, load_regex() exits */
      learner.retype |= (1<<load_regex(optarg));
      break;
    case 'i':
      m_options |= (1<<M_OPTION_I18N);
#if defined HAVE_LANGINFO_H
      if( !strcmp(nl_langinfo(CODESET), "UTF-8") ) {
	fprintf(stderr, "warning: you have UTF-8, so -i is not needed.\n");
      }
#endif
#if !defined HAVE_MBRTOWC
      fprintf(stderr, "warning: this tool was compiled without wide character support. Full internationalization is disabled.\n");
      m_options &= ~(1<<M_OPTION_I18N);
#endif
      break;
    default:
      break;
    }
  }

  /* end option processing */

  /* consistency checks */
  if( ((u_options>>U_OPTION_CLASSIFY) & 1) + 
      ((u_options>>U_OPTION_LEARN) & 1) != 1 ) {
    fprintf(stderr, 
	    "error: please use either -c or -l option.\n");
    usage(argv);
    exit(0);
  }

  if( (u_options & (1<<U_OPTION_DECIMATE)) &&
      !(u_options & (1<<U_OPTION_LEARN)) ) {
    fprintf(stderr,
	    "warning: option -x ignored, applies only when learning.\n");
    u_options &= ~(1<<U_OPTION_DECIMATE);
  }

  if( u_options & (1<<U_OPTION_DUMP) ) {
    if( u_options & (1<<U_OPTION_CLASSIFY) ) {
      u_options &= ~(1<<U_OPTION_DUMP); /* no sense */
    } else if( u_options & (1<<U_OPTION_VERBOSE) ) {
      u_options &= ~(1<<U_OPTION_VERBOSE); /* verbose writes garbage to stdout */
      u_options &= ~(1<<U_OPTION_DEBUG);
    }
  }

  if( !(m_options & (1<<M_OPTION_TEXT_FORMAT)) &&
      !(m_options & (1<<M_OPTION_MBOX_FORMAT)) &&
      !(m_options & (1<<M_OPTION_HTML)) &&
      !(m_options & (1<<M_OPTION_XML)) ) {
    m_options |= (1<<M_OPTION_TEXT_FORMAT);
  }

  if( ((m_options>>M_OPTION_TEXT_FORMAT) & 1) + 
      ((m_options>>M_OPTION_MBOX_FORMAT) & 1) > 1 ) {
    fprintf(stderr, 
	    "error: please use only one of either -T text or -T email options.\n");
    usage(argv);
    exit(0);
  }

  if( ((m_options>>M_OPTION_XML) & 1) + 
      ((m_options>>M_OPTION_HTML) & 1) > 1 ) {
    fprintf(stderr, 
	    "error: please use only one of either -T xml or -T html options.\n");
    usage(argv);
    exit(0);
  }

  /* mbox format needs html unless user wants xml */
  if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
    if( !(m_options & (1<<M_OPTION_XML)) ) {
      m_options |= (1<<M_OPTION_HTML);
    }
  }


  if( (u_options & (1<<U_OPTION_APPEND)) &&
      (u_options & (1<<U_OPTION_FILTER)) ) {
    u_options &= ~(1<<U_OPTION_APPEND);
    fprintf(stderr, 
	    "warning: disabling option -a, because it cannot be used with -f.\n");
  }

  /* decide if we need some options */

  if( !regex_count ) {
    m_options &= ~(1<<M_OPTION_USE_REGEX);
  }

  if( m_options & (1<<M_OPTION_MULTINOMIAL) ) { 
    if( cat_count == 1 ) {
      if( cat[0].model_type == simple ) {
	m_options |= (1<<M_OPTION_CALCENTROPY);
      }
    } else if( cat_count > 1 ) {
      for(c = 1; c < cat_count; c++) {
	if( cat[c].model_type == sequential ) { break; }
	if( cat[c].retype != cat[c-1].retype ) { break; }
      }
      if( c == cat_count ) {
	m_options |= (1<<M_OPTION_CALCENTROPY);
      }
    }
    if( !(m_options & (1<<M_OPTION_CALCENTROPY)) ) {
      fprintf(stderr,
	      "warning: not all categories support multinomial calculations, disabling (-M switch)\n");
    }
  }

  if( m_options & (1<<M_OPTION_I18N) ) {
#if defined HAVE_LIBBOOST_REGEX
#else
    if( u_options & (1<<U_OPTION_VERBOSE) ) {
      fprintf(stderr, 
	      "information: regexes operate in multibyte encoding.\n");
    }
#endif	
  }

  /* unless overridden, we use alpha character classes only */
  if( !(m_options & (1<<M_OPTION_CHAR_ALPHA)) &&
      !(m_options & (1<<M_OPTION_CHAR_ALNUM)) &&
      !(m_options & (1<<M_OPTION_CHAR_CEF)) &&
      !(m_options & (1<<M_OPTION_CHAR_ADP)) &&
      !(m_options & (1<<M_OPTION_CHAR_GRAPH)) ) {
    if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
      m_options |= (1<<M_OPTION_CHAR_ADP);
    } else {
      m_options |= (1<<M_OPTION_CHAR_ALPHA);
    }
  }

  if( !(u_options & (1<<U_OPTION_JAYNES)) &&
      !(u_options & (1<<U_OPTION_LAPLACE)) &&
      !(u_options & (1<<U_OPTION_DIRICHLET)) ) {
    if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
      u_options |= (1<<U_OPTION_LAPLACE);
    } else {
      u_options |= (1<<U_OPTION_DIRICHLET);
    }
  }

  /* set up callbacks */

  if( u_options & (1<<U_OPTION_CLASSIFY) ) {

    if( m_options & (1<<M_OPTION_CALCENTROPY) ) {
      init_empirical(&empirical, 
		     default_max_tokens, 
		     default_max_hash_bits); /* sets cached to zero */
    }

    if( cat_count == 1 ) {
      /* single category */

      preprocess_fun = NULL;
      word_fun = score_word;
      if( u_options & (1<<U_OPTION_FILTER) ) {
	u_options |= (1<<U_OPTION_FASTEMP);
	empirical.track_features = 1; 
	post_line_fun = line_score_single_category;
	postprocess_fun = NULL;
      } else {
	post_line_fun = NULL;
	postprocess_fun = score_single_category;
      }

    } else {
      /* multiple categories */

      preprocess_fun = NULL;
      word_fun = score_word;
      if( u_options & (1<<U_OPTION_FILTER) ) {
	u_options |= (1<<U_OPTION_FASTEMP);
	empirical.track_features = 1; 
	post_line_fun = line_score_multiple_categories;
	postprocess_fun = NULL;
      } else {
	post_line_fun = NULL;
	postprocess_fun = score_multiple_categories;
      }

    }
  } else if( u_options & (1<<U_OPTION_LEARN) ) {
    /* category learning */

    preprocess_fun = init_learner;
    word_fun = hash_word_and_learn;
    if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
      post_line_fun = count_mbox_messages;
    } else {
      post_line_fun = NULL;
    }
    postprocess_fun = optimize_learner_and_save;

  } else { /* something wrong ? */
    usage(argv);
    exit(0);
  }

  /* handles some common filtering options */
  if( (u_options & (1<<U_OPTION_INDENTED)) ||
      (u_options & (1<<U_OPTION_APPEND)) ) {
    pre_line_fun = handle_indents_and_appends;
  }

  if( preprocess_fun ) { (*preprocess_fun)(); }


  init_file_handling();

  /* now process each file on the command line,
     or if none provided read stdin */
  while( (optind > -1) && *(argv + optind) && !(cmd & (1<<CMD_QUITNOW)) ) {
    /* if it's a filename, process it */
    if( (input = fopen(argv[optind], "r")) ) {

      if( in_iobuf ) {
	/* choose appropriate buffering mode */
	if( fstat(fileno(input), &statinfo) == 0 ) {
	  switch(statinfo.st_mode & S_IFMT) {
	  case S_IFREG:
	  case S_IFBLK:
	    setvbuf(input, in_iobuf, _IOFBF, BUFFER_MAG * system_pagesize);
	    break;
	  case S_IFSOCK:
	  case S_IFIFO:
	  case S_IFCHR:
	    setvbuf(input, NULL, 
		    (u_options & (1<<U_OPTION_FILTER)) ? _IOLBF : _IOFBF, 
		    BUFFER_MAG * system_pagesize);
	    break;
	  case S_IFDIR:
	  case S_IFLNK:
	    /* nothing */
	    break;
	  }

	}
      }

      u_options |= (1<<U_OPTION_STDIN);

      if( (u_options & (1<<U_OPTION_VERBOSE)) && 
	  !(u_options & (1<<U_OPTION_CLASSIFY))) {
	fprintf(stdout, "processing file %s\n", argv[optind]);
      }

      /* set some initial options */
      reset_xml_character_filter(xmlRESET);

      if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
	reset_mbox_line_filter();
      }

      if( !(m_options & (1<<M_OPTION_I18N)) ) {
	process_file(input, 
		     ((m_options & (1<<M_OPTION_MBOX_FORMAT)) ? 
		      mbox_line_filter : NULL), 
		     (((m_options & (1<<M_OPTION_XML)) || 
		       (m_options & (1<<M_OPTION_HTML))) ? 
		      xml_character_filter : NULL), 
		     word_fun, pre_line_fun, post_line_fun);
      } else {
#if defined HAVE_MBRTOWC
	w_process_file(input, 
		       ((m_options & (1<<M_OPTION_MBOX_FORMAT)) ? 
			w_mbox_line_filter : NULL), 
		       (((m_options & (1<<M_OPTION_XML)) || 
			 (m_options & (1<<M_OPTION_HTML))) ? 
			w_xml_character_filter : NULL), 
		       word_fun, pre_line_fun, post_line_fun);
#else
	fprintf(stderr, "error: international support not available (recompile).\n");
#endif
      }

      fclose(input);

    } else { /* unrecognized file name */

      fprintf(stderr, "error: couldn't open %s\n", argv[optind]);
      usage(argv);
      exit(0);

    }
    optind++;
  }
  /* in case no files were specified, get input from stdin */
  if( !(u_options & (1<<U_OPTION_STDIN)) ) {

    if( (u_options & (1<<U_OPTION_VERBOSE)) && 
	!(u_options & (1<<U_OPTION_CLASSIFY)) ) {
      fprintf(stdout, "taking input from stdin\n");
    }

    if( in_iobuf ) {
      /* choose appropriate buffering mode */
      if( fstat(fileno(stdin), &statinfo) == 0 ) {
	switch(statinfo.st_mode & S_IFMT) {
	case S_IFREG:
	case S_IFBLK:
	  setvbuf(stdin, in_iobuf, _IOFBF, BUFFER_MAG * system_pagesize);
	  break;
	case S_IFSOCK:
	case S_IFIFO:
	case S_IFCHR:
	  setvbuf(stdin, NULL, 
		  (u_options & (1<<U_OPTION_FILTER)) ? _IOLBF : _IOFBF, 
		  BUFFER_MAG * system_pagesize);
	  break;
	case S_IFDIR:
	case S_IFLNK:
	  /* nothing */
	  break;
	}

      }
    }

    /* set some initial options */
    reset_xml_character_filter(xmlRESET);

    if( m_options & (1<<M_OPTION_MBOX_FORMAT) ) {
      reset_mbox_line_filter();
    }

    if( !(m_options & (1<<M_OPTION_I18N)) ) {
      process_file(stdin, 
		   ((m_options & (1<<M_OPTION_MBOX_FORMAT)) ? 
		    mbox_line_filter : NULL), 
		   (((m_options & (1<<M_OPTION_XML)) || 
		     (m_options & (1<<M_OPTION_HTML))) ? 
		    xml_character_filter : NULL), 
		   word_fun, pre_line_fun, post_line_fun);
    } else {
#if defined HAVE_MBRTOWC
      w_process_file(stdin, 
		     ((m_options & (1<<M_OPTION_MBOX_FORMAT)) ? 
		      w_mbox_line_filter : NULL), 
		     (((m_options & (1<<M_OPTION_XML)) || 
		       (m_options & (1<<M_OPTION_HTML))) ? 
		      w_xml_character_filter : NULL), 
		     word_fun, pre_line_fun, post_line_fun);
#else
      fprintf(stderr, "error: international support not available (recompile).\n");
#endif
    }

    /* must close before freeing in_iobuf, in case setvbuf was called */
    fclose(stdin); 

  }
  
  if( postprocess_fun ) { (*postprocess_fun)(); }

  cleanup_file_handling();


#undef GOODGUY
#if defined GOODGUY
  /* normally we should free everything nicely, but there's no
     point, since the kernel will free the process memory. It's actually
     faster to not free... */

  for(c = 0; c < category_count; k++) {
    free_category(&cat[c]);
  }
  free_empirical(&empirical);

  free_all_regexes();

#endif

  cleanup_signal_handling();

  exit(exit_code);
}


